// TODO: check all `db.get()` and similar calls recieve a proper type argument (aka add `.toString()`, etc. to those calls)
// TODO: can we rely on db having "foreign keys"? When item is deleted will automaticly be all relations to it unset?
//       Similarly, will saving item also save all its related items no-yet-saved in db, or do they need to saved individually?
//       Also, is the assumption that `db.save(MyType, {myProperty: undefined})` unsets value in db correct?

import { SubstrateEvent } from '@dzlzv/hydra-common'
import { DatabaseManager } from '@dzlzv/hydra-db-utils'
import ISO6391 from 'iso-639-1';
import BN from 'bn.js'
import { u64 } from '@polkadot/types/primitive';

// protobuf definitions
import {
  ChannelMetadata,
  ChannelCategoryMetadata,
  PublishedBeforeJoystream as PublishedBeforeJoystreamMetadata,
  License as LicenseMetadata,
  MediaType as MediaTypeMetadata,
  VideoMetadata,
  VideoCategoryMetadata,
} from '@joystream/content-metadata-protobuf'

import {
  Content,
} from '../../../generated/types'

import {
  inconsistentState,
  prepareDataObject,
} from '../common'

// primary entities
import { CuratorGroup } from 'query-node/src/modules/curator-group/curator-group.model'
import { Channel } from 'query-node/src/modules/channel/channel.model'
import { ChannelCategory } from 'query-node/src/modules/channel-category/channel-category.model'
import { Video } from 'query-node/src/modules/video/video.model'
import { VideoCategory } from 'query-node/src/modules/video-category/video-category.model'

// secondary entities
import { Language } from 'query-node/src/modules/language/language.model'
import { License } from 'query-node/src/modules/license/license.model'
import { VideoMediaEncoding } from 'query-node/src/modules/video-media-encoding/video-media-encoding.model'
import { VideoMediaMetadata } from 'query-node/src/modules/video-media-metadata/video-media-metadata.model'

// Asset
import { AssetAvailability } from 'query-node/src/modules/enums/enums';
import {
  DataObject,
  LiaisonJudgement
} from 'query-node/src/modules/data-object/data-object.model'
import {
  DataObjectOwner,
  DataObjectOwnerMember,
} from 'query-node/src/modules/variants/variants.model'

// Joystream types
import {
  ContentParameters,
  NewAsset,
  ContentActor,
} from '@joystream/types/augment'

/*
  Asset either stored in storage or describing list of URLs.
*/
type AssetStorageOrUrls = DataObject | string[]

/*
  Type guard differentiating asset stored in storage from asset describing a list of URLs.
*/
function isAssetInStorage(dataObject: AssetStorageOrUrls): dataObject is DataObject {
  if (Array.isArray(dataObject)) {
    return false
  }

  return true
}

/*
  Reads information from the event and protobuf metadata and constructs changeset that's fit to be used when saving to db.
*/
export async function readProtobuf(
  type: Channel | ChannelCategory | Video | VideoCategory,
  metadata: Uint8Array,
  assets: NewAsset[], // assets provided in event
  db: DatabaseManager,
  blockNumber: number,
  actor: ContentActor,
): Promise<Partial<typeof type>> {
  // process channel
  if (type instanceof Channel) {
    const meta = ChannelMetadata.deserializeBinary(metadata)
    const metaAsObject = meta.toObject()
    const result = metaAsObject as any as Channel

    // prepare cover photo asset if needed
    if (metaAsObject.coverPhoto !== undefined) {
      const asset = await extractAsset(metaAsObject.coverPhoto, assets, db, blockNumber, actor)
      integrateAsset('coverPhoto', result, asset) // changes `result` inline!
      delete metaAsObject.coverPhoto
    }

    // prepare avatar photo asset if needed
    if (metaAsObject.avatarPhoto !== undefined) {
      const asset = await extractAsset(metaAsObject.avatarPhoto, assets, db, blockNumber, actor)
      integrateAsset('avatarPhoto', result, asset) // changes `result` inline!
      delete metaAsObject.avatarPhoto
    }

    // prepare language if needed
    if (metaAsObject.language) {
      result.language = await prepareLanguage(metaAsObject.language, db)
    }

    return result
  }

  // process channel category
  if (type instanceof ChannelCategory) {
    return ChannelCategoryMetadata.deserializeBinary(metadata).toObject()
  }

  // process video
  if (type instanceof Video) {
    const meta = VideoMetadata.deserializeBinary(metadata)
    const metaAsObject = meta.toObject()
    const result = metaAsObject as any as Video

    // prepare video category if needed
    if (metaAsObject.category !== undefined) {
      result.category = await prepareVideoCategory(metaAsObject.category, db)
    }

    // prepare media meta information if needed
    if (metaAsObject.mediaType) {
      // prepare video file size if poosible
      const videoSize = await extractVideoSize(assets, metaAsObject.video)

      result.mediaMetadata = await prepareVideoMetadata(metaAsObject, videoSize)
      delete metaAsObject.mediaType
    }

    // prepare license if needed
    if (metaAsObject.license) {
      result.license = await prepareLicense(metaAsObject.license)
    }

    // prepare thumbnail photo asset if needed
    if (metaAsObject.thumbnailPhoto !== undefined) {
      const asset = await extractAsset(metaAsObject.thumbnailPhoto, assets, db, blockNumber, actor)
      integrateAsset('thumbnail', result, asset) // changes `result` inline!
      delete metaAsObject.thumbnailPhoto
    }

    // prepare video asset if needed
    if (metaAsObject.video !== undefined) {
      const asset = await extractAsset(metaAsObject.video, assets, db, blockNumber, actor)
      integrateAsset('media', result, asset) // changes `result` inline!
      delete metaAsObject.video
    }

    // prepare language if needed
    if (metaAsObject.language) {
      result.language = await prepareLanguage(metaAsObject.language, db)
    }

    // prepare information about media published somewhere else before Joystream if needed.
    if (metaAsObject.publishedBeforeJoystream) {
      // this will change the `channel`!
      handlePublishedBeforeJoystream(result, metaAsObject.publishedBeforeJoystream.date)
    }

    return result
  }

  // process video category
  if (type instanceof VideoCategory) {
    return VideoCategoryMetadata.deserializeBinary(metadata).toObject()
  }

  // this should never happen
  throw `Not implemented type: ${type}`
}

export function convertContentActorToOwner(actor: ContentActor): typeof DataObjectOwner {
  if (actor.isMember) {
    const owner = new DataObjectOwnerMember()
    owner.member = actor.asMember.toBn()

    return owner
  }

  if (actor.isLead) {
    // TODO: who will be owner?
  }

  if (actor.isCurator) {
    // TODO: who will be owner?
  }

  throw 'Not-implemented ContentActor type used'
}

function handlePublishedBeforeJoystream(video: Video, publishedAtString?: string) {
  // published elsewhere before Joystream
  if (publishedAtString) {
    video.publishedBeforeJoystream = new Date(publishedAtString)
  }

  // unset publish info
  video.publishedBeforeJoystream = undefined // plan deletion (will have effect when saved to db)
}

/*
  Converts event asset into data object or list of URLs fit to be saved to db.
*/
async function convertAsset(rawAsset: NewAsset, db: DatabaseManager, blockNumber: number, actor: ContentActor): Promise<AssetStorageOrUrls> {
  // is asset describing list of URLs?
  if (rawAsset.isUrls) {
    const urls = rawAsset.asUrls.toArray().map(item => item.toString())

    return urls
  }

  // !rawAsset.isUrls && rawAsset.isUpload // asset is in storage

  // prepare data object
  const contentParameters: ContentParameters = rawAsset.asUpload
  const owner = convertContentActorToOwner(actor)
  const dataObject = await prepareDataObject(contentParameters, blockNumber, owner)


  return dataObject
}

/*
  Selects asset from provided set of assets and prepares asset data fit to be saved to db.
*/
async function extractAsset(
  assetIndex: number,
  assets: NewAsset[],
  db: DatabaseManager,
  blockNumber: number,
  actor: ContentActor,
): Promise<AssetStorageOrUrls> {
  // ensure asset index is valid
  if (assetIndex > assets.length) {
    return inconsistentState(`Non-existing asset extraction requested`, {assetsProvided: assets.length, assetIndex})
  }

  // convert asset to data object record
  return convertAsset(assets[assetIndex], db, blockNumber, actor)
}

/*
  As a temporary messure to overcome yet-to-be-implemented features in Hydra, we are using redudant information
  to describe asset state. This function introduces all redudant data needed to be saved to db.

  Changes `result` argument!
*/
function integrateAsset<T>(propertyName: string, result: Object, asset: AssetStorageOrUrls) {
  // helpers - property names
  const nameUrl = propertyName + 'Urls'
  const nameDataObject = propertyName + 'DataObject'
  const nameAvailability = propertyName + 'Availability'

  // is asset saved in storage?
  if (!isAssetInStorage(asset)) {
    // (un)set asset's properties
    result[nameUrl] = asset
    result[nameAvailability] = AssetAvailability.ACCEPTED
    result[nameDataObject] = undefined // plan deletion (will have effect when saved to db)

    return result
  }

  // prepare conversion table between liaison judgment and asset availability
  const conversionTable = {
    [LiaisonJudgement.ACCEPTED]: AssetAvailability.ACCEPTED,
    [LiaisonJudgement.PENDING]: AssetAvailability.PENDING,
    [LiaisonJudgement.REJECTED]: AssetAvailability.INVALID,
  }

  // (un)set asset's properties
  result[nameUrl] = undefined // plan deletion (will have effect when saved to db)
  result[nameAvailability] = conversionTable[asset.liaisonJudgement]
  result[nameDataObject] = asset
}

async function extractVideoSize(assets: NewAsset[], assetIndex: number | undefined): Promise<BN | undefined> {
  // escape if no asset is required
  if (assetIndex === undefined) {
    return undefined
  }

  // ensure asset index is valid
  if (assetIndex > assets.length) {
    return inconsistentState(`Non-existing asset video size extraction requested`, {assetsProvided: assets.length, assetIndex})
  }

  const rawAsset = assets[assetIndex]

  // escape if asset is describing URLs (can't get size)
  if (rawAsset.isUrls) {
    return undefined
  }

  // !rawAsset.isUrls && rawAsset.isUpload // asset is in storage

  // extract video size
  const contentParameters: ContentParameters = rawAsset.asUpload
  // `size` is masked by `size` special name in struct that's why there needs to be `.get('size') as u64`
  const videoSize = (contentParameters.get('size') as unknown as u64).toBn()

  return videoSize
}

async function prepareLanguage(languageIso: string, db: DatabaseManager): Promise<Language> {
  // validate language string
  const isValidIso = ISO6391.validate(languageIso);

  // ensure language string is valid
  if (!isValidIso) {
    return inconsistentState(`Invalid language ISO-639-1 provided`, languageIso)
  }

  // load language
  const language = await db.get(Language, { where: { iso: languageIso }})

  // return existing language if any
  if (language) {
    return language;
  }

  // create new language
  const newLanguage = new Language({
    iso: languageIso
  })

  return newLanguage
}

async function prepareLicense(licenseProtobuf: LicenseMetadata.AsObject): Promise<License> {
  // NOTE: Deletion of any previous license should take place in appropriate event handling function
  //       and not here even it might appear so.

  // crete new license
  const license = new License(licenseProtobuf)

  return license
}

async function prepareVideoMetadata(videoProtobuf: VideoMetadata.AsObject, videoSize: BN | undefined): Promise<VideoMediaMetadata> {
  // create new encoding info
  const encoding = new VideoMediaEncoding(videoProtobuf.mediaType)

  // create new video metadata
  const videoMeta = new VideoMediaMetadata({
    encoding,
    pixelWidth: videoProtobuf.mediaPixelWidth,
    pixelHeight: videoProtobuf.mediaPixelHeight,
  })

  // fill in video size if provided
  if (videoSize !== undefined) {
    videoMeta.size = videoSize
  }

  return videoMeta
}

async function prepareVideoCategory(categoryId: number, db: DatabaseManager): Promise<VideoCategory> {
  // load video category
  const category = await db.get(VideoCategory, { where: { id: categoryId }})

  // ensure video category exists
  if (!category) {
    return inconsistentState('Non-existing video category association with video requested', categoryId)
  }

  return category
}
